
#include <iostream>
#include <vector>
#include <string>

#include <boost/program_options.hpp>
#include <boost/thread.hpp>
#include <boost/asio.hpp>

#include "client.hpp"
#include "server.hpp"

namespace po = boost::program_options;
namespace io = boost::asio;

std::map<int, boost::posix_time::ptime> start_times;
std::map<int, boost::posix_time::ptime> end_times;

namespace {

void wait_q();
void print_options();
void print_times();

}

int main(int argc, char* argv[])
{
    po::options_description desc("Allowed options");

    int num_sockets;
    int num_packets;
    size_t packet_size;
    unsigned short local_port;
    std::string local_address;
    unsigned short remote_port;
    std::string remote_address;
    std::string log_path;

    desc.add_options()
    ("help", "produce help message")
    ("server", "act as a server")
    ("client", "act as a client")
    ("header", "print tcp info options header")
    ("localport", po::value<unsigned short>(&local_port)->default_value(10000), "local port (only for server)")
    ("localaddress", po::value<std::string>(&local_address)->default_value("0.0.0.0"), "local address (only for server)")
    ("remoteport", po::value<unsigned short>(&remote_port)->default_value(10000), "remote port (only for client)")
    ("remoteaddress", po::value<std::string>(&remote_address), "remote address (only for client)")
    ("connections", po::value<int>(&num_sockets)->default_value(1), "number of client connections")
    ("packets", po::value<int>(&num_packets)->default_value(1), "number of sent packets per connection")
    ("size", po::value<size_t>(&packet_size)->default_value(64), "size of the packets");

    po::variables_map vm;

    try
    {
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);
    }
    catch (...)
    {
        std::cerr << desc << std::endl;

        return -1;
    }

    if (vm.count("help"))
    {
        std::cout << desc << std::endl;

        return -1;
    }

    if (vm.count("server") && vm.count("client"))
    {
        std::cerr << desc << std::endl;

        return -1;
    }

    if (vm.count("header"))
    {
        print_options();

        return 0;
    }

    if (vm.count("server"))
    {
        boost::thread th(start_server, boost::cref(local_address), local_port);
        wait_q();
	print_times();
    }
    else if (vm.count("client"))
    {
        if (remote_address.empty())
        {
            std::cerr << desc << std::endl;

            return -1;
        }

        start_client(remote_address, remote_port, num_sockets, num_packets, packet_size);
    }
    else
    {
        std::cerr << desc << std::endl;

        return -1;
    }
}

namespace {

void wait_q()
{
    while (true)
    {
        std::cout << "Digit q to exit" << std::endl;
        char c;
        std::cin >> c;

        if (c == 'q')
        {
            break;
        }
    }

    std::cout << "Exiting..." << std::endl;
}

void print_options()
{
    std::cout
        << "[Time]"
        << " [retransmits]"
        << " [backoff]"
        << " [lost]"
        << " [retrans]"
        << " [total_retrans]"
        << " [last_data_sent]"
        << " [last_ack_sent]"
        << " [rcv_ssthresh]"
        << " [rtt]"
        << " [rttvar]"
        << " [snd_ssthresh]"
        << " [snd_cwnd]"
        << " [reordering]"
        << std::endl;
}

void print_times()
{
    std::cerr << "start times size = " << start_times.size() << " - end_times size = " << end_times.size()<< std::endl;
    
    typedef std::map<int, boost::posix_time::ptime>::const_iterator Iterator;
    double tot_times = 0;
    double num = 0;

    for (Iterator i = start_times.begin(); i!= start_times.end(); ++i)
    {
        Iterator e = end_times.find(i->first);

        if (e == end_times.end())
        {
            continue;
        }

        tot_times += (e->second - i->second).total_microseconds();
        ++num;
    }

    std::cout << "Average time: " << (tot_times / num) << " calculated on " << num << " samples" << std::endl;
}

}
